use regex::Regex;
use std::num::ParseIntError;
use std::str::FromStr;
use thiserror::Error;

/// Chunk error
#[derive(Debug, Error)]
pub enum ChunkCoordinateErr {
    #[error("Failed to parse int: {0}")]
    ParseIntError(#[from] ParseIntError),
    #[error("Regex error: {0}")]
    RegexError(#[from] regex::Error),
    #[error("Invalid chunk coordinate")]
    InvalidChunk,
}

/// Chunk Coordinate pair
#[derive(Debug)]
pub struct ChunkCoordinate {
    /// X Coordinate
    pub x: i32,
    /// Z Coordinate
    pub z: i32,
}

impl FromStr for ChunkCoordinate {
    type Err = ChunkCoordinateErr;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let re = Regex::new(r"\((?P<x>-?[0-9]*),(?P<z>-?[0-9]*)\)")?;

        if let Some(cap) = re.captures(s) {
            let x = cap["x"].parse::<i32>()?;
            let z = cap["z"].parse::<i32>()?;

            Ok(Self { x, z })
        } else {
            Err(ChunkCoordinateErr::InvalidChunk)
        }
    }
}

#[cfg(test)]
mod test {
    use crate::chunk_coordinate::ChunkCoordinate;
    use std::str::FromStr;

    #[test]
    fn parse_chunk_coordinate_success() {
        let chunk_x = 5;
        let chunk_z = -15;
        let chunk_string = format!("({chunk_x},{chunk_z})");

        let chunk = ChunkCoordinate::from_str(chunk_string.as_str()).unwrap();

        assert_eq!(chunk.x, chunk_x);
        assert_eq!(chunk.z, chunk_z);
    }

    #[test]
    fn parse_chunk_coordinate_failure() {
        assert!(ChunkCoordinate::from_str("(nan,5)").is_err());
    }
}
